/****************************************************************************/
/*               SPI FlashROM programmer on PC parallel port                */
/****************************************************************************/
/* Created: 24.3.2008 [DJGPP]                                               */
/* Last modified: 28.10.2008 [DJGPP]                                        */
/* Copyright (C) 2008 by RayeR/Martin Rehak                                 */
/* Contact: rayer^seznam*cz                                                 */
/****************************************************************************/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <time.h>

#define SPI_CS_PIN 5                   // D5 (CS#)
#define SPI_SCK_PIN 6                  // D6 (CLK)
#define SPI_MOSI_PIN 7                 // D7 (DIO)
#define SPI_MISO_PIN 6                 // ACK (DO)
#define SPI_BUFFER_SIZE 100000LU       // 100kB SPI data buffer (64kB+rezerva)

//***************** SPI lines control macros ********************************
#define SPI_SET_CS outportb(lpt_base,inportb(lpt_base)|(1<<SPI_CS_PIN))
#define SPI_CLR_CS outportb(lpt_base,inportb(lpt_base)&(~(1<<SPI_CS_PIN)))
#define SPI_SET_SCK outportb(lpt_base,inportb(lpt_base)|(1<<SPI_SCK_PIN))
#define SPI_CLR_SCK outportb(lpt_base,inportb(lpt_base)&(~(1<<SPI_SCK_PIN)))
#define SPI_SET_MOSI outportb(lpt_base,inportb(lpt_base)|(1<<SPI_MOSI_PIN))
#define SPI_CLR_MOSI outportb(lpt_base,inportb(lpt_base)&(~(1<<SPI_MOSI_PIN)))
#define SPI_GET_MISO ((inportb(lpt_base+1)>>SPI_MISO_PIN)&1)

//***************** SPI FlashROM command codes ******************************
// 25Xxx FlashROM description:
// SPI mode 0: SCK idle state = low, DIO sampled on rising edge, DO clocked on falling edge, MSB first
// 512kB memory, total 8 blocks per 64 kB, 1 block = 16 sectors per 4 kB, 1 sector = 16 pages per 256B
#define SPI_FLASH_WREN         0x06    // povol zapis
#define SPI_FLASH_WRDIS        0x04    // zakaz zapis
#define SPI_FLASH_RDSTAT       0x05    // cti status registr; S7:0 - bit7=stat reg protect, bit5=top/bot write protect, bit4:2=block protect bits, bit1=WE latch, bit0=busy
#define SPI_FLASH_WRSTAT       0x01    // zapis status registr; S7:0
#define SPI_FLASH_RDDATA       0x03    // cti data; A23:16, A15:A8, A7:0, D7:0, next DB till the end
#define SPI_FLASH_FASTRD       0x0B    // rychle cteni; A23:16, A15:A8, A7:0, dummyB, D7:0, next DB till the end
#define SPI_FLASH_FASTRD2      0x3B    // rychle cteni 2 linkami DO a DIO; A23:16, A15:A8, A7:0, dummyB, D7:0 bit interleaved, next DB till the end
#define SPI_FLASH_PAGEPGM      0x02    // programovani stranky; A23:16, A15:A8, A7:0, D7:0, next DB till the end
#define SPI_FLASH_BLKERASE     0xD8    // smazani bloku (64kB pro 25Xxx); A23:16, A15:A8, A7:0
#define SPI_FLASH_SECERASE     0x20    // smazani sektoru (4kB pro 25Xxx); A23:16, A15:A8, A7:0
#define SPI_FLASH_CHIPERASE    0xC7    // smazani cele FlashROM
#define SPI_FLASH_PWRDWN       0xB9    // uspani
#define SPI_FLASH_RELPWRDWN    0xAB    // probuzeni z uspani; dummyB, dummyB, dummyB, ID7:0 (10h=X10, 11h=X20, 12h=X40, 13h=X80)
#define SPI_FLASH_MIDDID       0x90    // ID vyrobce a zarizeni; dummyB, dummyB, 0, M7:0 (EFh=Winbond), ID7:0 (10h=X10, 11h=X20, 12h=X40, 13h=X80)
#define SPI_FLASH_JEDECID      0x9F    // JEDEC ID vyrobce; M7:0, ID15:8, ID7:0  (3011h=X10, 3012h=X20, 3013h=X40, 3014h=X80)

#define SPI_FLASH_JID_ATMEL    0x1F    // Atmel
#define SPI_FLASH_JID_STM      0x20    // ST Microelectronic
#define SPI_FLASH_JID_SST      0xBF    // SST (Silicon Storage Technology)
#define SPI_FLASH_JID_MACRONIX 0xC2    // Macronix
#define SPI_FLASH_JID_WINBOND  0xEF    // Winbond

#define SPI_FLASH_PAGE_SIZE    256     // velikost stranky, max. jednotka pro zapis
#define SPI_FLASH_SECTOR_SIZE  4096    // velikost sektoru, min. jednotka pro mazani
#define SPI_FLASH_BLOCK_SIZE   65536   // velikost bloku, vetsi jednotka pro mazani

#define Word unsigned short
#define DWord unsigned long
#define Byte unsigned char
#define udelay usleep

Word lpt_base=0x378;                   // LPT1 I/O base address
DWord spi_sck_pulse_delay=0;           // pridavna prodleva k dobe trvani SCK pulsu
Byte spi_di_buffer[SPI_BUFFER_SIZE];   // SPI data input buffer
Byte spi_do_buffer[SPI_BUFFER_SIZE];   // SPI data output buffer
DWord spi_flash_size=0;                // celkova velikost FlashROM

//***************** nastavi SPI linky do klidovych urovni *******************
void spi_init(void)
{
  SPI_SET_CS;                          // CS# = high
  SPI_CLR_SCK;                         // CLK = low
  SPI_CLR_MOSI;                        // MOSI = low
  memset(spi_di_buffer,0,SPI_BUFFER_SIZE); // clear input data buffer
  memset(spi_do_buffer,0,SPI_BUFFER_SIZE); // clear output data buffer
}

//***************** vypis hexadecimale a textove dany pocet Bytu z SPI bufferu
void spi_print_buffer(Byte *p_buffer, long count, long offset)
{
  long i, j;
  for (i=0; i<count; i+=16)            // pres vsechny 16-Bytove bloky
    {
    printf("%06lX: ",i+offset);        // vypis adresu
    for (j=0; j<16; j++)               // pres vsechny byte v 16B bloku
      {
      if ((j&0x7)==0)                  // na zacatku a po 8 Bytech
        putchar(' ');                  // oddelovaci mezera
      if ((j+i)<count)                 // pokud nejsou vypsany vsechny Byte
        printf("%02X ",p_buffer[j+i]); // vypis 1 Byte v hexa
      else
        printf("   ");                 // jinak misto nej vypis mezery
      }
    putchar(' ');                      // oddelovaci mezera
    for (j=0; j<16; j++)               // pres vsechny byte v 16B bloku
      if ((j+i)<count)                 // pokud nejsou vypsany vsechny Byte
        {
        if (p_buffer[j+i]>=' ')        // pokud je to tisknutelny znak
          putchar(p_buffer[j+i]);      // vypis ho
        else
          putchar(' ');                // jinak vypis mezeru
        }
    putchar('\n');                     // konec radky
    }
}

//***************** zapise a precte 1 Byte po SPI bez obsluhy CS# signalu ***
Byte spi_rw_byte_no_cs(Byte data_out)
{
  Byte data_in=0, i;
  for (i=0; i<8; i++)                  // pres vsechny bity Bytu
    {
    data_in<<=1;                       // shift to next input bit, will be placed at bit0
    if ((data_out&0x80)==0)            // start from MSB to LSB
      SPI_CLR_MOSI;                    // data bit is 0
    else
      SPI_SET_MOSI;                    // data bit is 1
    udelay(spi_sck_pulse_delay);       // data setup delay
    SPI_SET_SCK;                       // slave samples data on rising clock edge
    data_in|=SPI_GET_MISO;             // sample slave output, from MSB to LSB
    udelay(spi_sck_pulse_delay);       // clock pulse delay
    SPI_CLR_SCK;                       // clock back to default state
    data_out<<=1;                      // shift to next output bit, MSB will be send
    }
  return(data_in);                     // return readed Byte
}

//***************** zapise a precte 1 Byte po SPI vcetne rizeni CS# *********
Byte spi_rw_byte(Byte data_out)
{
  Byte data_in;
  SPI_CLR_SCK;                         // clock default state
  SPI_CLR_CS;                          // chip select enabled
  udelay(spi_sck_pulse_delay);         // wait for chip active
  data_in=spi_rw_byte_no_cs(data_out); // write/read Byte
  udelay(spi_sck_pulse_delay);         // wait for chip inactive
  SPI_SET_CS;                          // chip select disabled
  return(data_in);
}

//***************** zapise a precte blok dat po SPI vcetne rizeni CS# *******
void spi_rw_block(Byte *p_do_buffer, Byte *p_di_buffer, DWord blk_size)
{
  long i;
  SPI_CLR_SCK;                         // clock default state
  SPI_CLR_CS;                          // chip select enabled
  udelay(spi_sck_pulse_delay);         // wait for chip active
  for (i=0; i<blk_size; i++)           // pres vsechny Byte bloku
    p_di_buffer[i]=spi_rw_byte_no_cs(p_do_buffer[i]); // write/read Byte
  udelay(spi_sck_pulse_delay);         // wait for chip inactive
  SPI_SET_CS;                          // chip select disabled
}

//***************** precte stavovy registr SPI FlashROM (bit0 = busy) *******
Byte spi_flash_status(void)
{
  spi_do_buffer[0]=SPI_FLASH_RDSTAT;   // put command code to output buffer
  spi_rw_block(spi_do_buffer,spi_di_buffer,2); // execute block read/write
  return(spi_di_buffer[1]);            // vrat 2. precteny Byte - status registr
}

//***************** ceka dokud je SPI FlashROM busy (napr. po mazani) *******
void spi_flash_wait_busy(int timeout)  // timeout v milisekundach
{
  while ((timeout--)>0)                // dokud nevyprsel timeout
    {
    udelay(1000);
    if ((spi_flash_status()&0x1)==0)   // cti status
      break;                           // je-li bit busy = 0, skonci
    }
}

//***************** povoli zapis do SPI FlashROM ****************************
#define spi_flash_write_enable() spi_rw_byte(SPI_FLASH_WREN)

//***************** zakaze zapis do SPI FlashROM ****************************
#define spi_flash_write_disable() spi_rw_byte(SPI_FLASH_WRDIS)

//***************** precte manufacturer-MSB a device ID-LSB SPI FlashROM ****
Word spi_flash_read_id(void)
{
  memset(spi_do_buffer,0,6);           // clear output buffer
  spi_do_buffer[0]=SPI_FLASH_MIDDID;   // put command code to output buffer
  spi_rw_block(spi_do_buffer,spi_di_buffer,6); // execute block read/write
  return(((Word)spi_di_buffer[4]<<8)|spi_di_buffer[5]);
}

//***************** precte JEDEC ID-Byte2, MemType-Byte1, MemSize-Byte0 *****
DWord spi_flash_read_jid(void)
{
  memset(spi_do_buffer,0,4);           // clear output buffer
  spi_do_buffer[0]=SPI_FLASH_JEDECID;  // put command code to output buffer
  spi_rw_block(spi_do_buffer,spi_di_buffer,4); // execute block read/write
  return(((DWord)spi_di_buffer[1]<<16)|((DWord)spi_di_buffer[2]<<8)|spi_di_buffer[3]);
}

//***************** precte blok dat od dane adresy z SPI FlashROM (eliminuje uvodni dummy Byte, data v bufferu zacinaji od 0)
void spi_flash_read_block(DWord start_address, DWord blk_size)
{
  long i;
  if (blk_size>(SPI_BUFFER_SIZE-4))
    {
    printf("ERROR: block read size (%luB) exceeds SPI read buffer (%luB)\n",blk_size,SPI_BUFFER_SIZE);
    return;
    }
  spi_do_buffer[0]=SPI_FLASH_RDDATA;   // put command code to output buffer
  spi_do_buffer[1]=(start_address>>16)&0xFF; // put A23:16 to output buffer
  spi_do_buffer[2]=(start_address>>8)&0xFF; // put A15:8 to output buffer
  spi_do_buffer[3]=(start_address>>0)&0xFF; // put A7:0 to output buffer
  spi_rw_block(spi_do_buffer,spi_di_buffer,blk_size+4); // execute block read/write
  for (i=0; i<blk_size; i++)           // pres vsechny Byty bloku
    spi_di_buffer[i]=spi_di_buffer[i+4]; // posun data o 4 Byte dolu
}

//***************** precte obsah cele SPI FlashROM do souboru ***************
int spi_flash_dump(char *file_name, DWord total_size)
{
  DWord address;
  FILE *fw=fopen(file_name,"wb");      // otevri soubor pro zapis
  if (fw==NULL)                        // pokud chyba, tak skonci
    {
    printf("ERROR: cannot create file %s\n",file_name);
    return(-1);
    }
  for (address=0; address<total_size; address+=SPI_FLASH_SECTOR_SIZE)
    {                                  // pres vsechny sektory
    printf("Reading sector at: %08lXh (%lu%%)\r",address,100*(address+SPI_FLASH_SECTOR_SIZE)/total_size);
    spi_flash_read_block(address,SPI_FLASH_SECTOR_SIZE); // nacti sektor do bufferu
    if (fwrite(spi_di_buffer,1,SPI_FLASH_SECTOR_SIZE,fw)!=SPI_FLASH_SECTOR_SIZE) // a zapis do souboru
      {
      printf("ERROR: cannot write sector %lu to file\n",address/SPI_FLASH_SECTOR_SIZE);
      return(-2);
      }
    }
  fclose(fw);                          // zavri soubor
  putchar('\n');
  return(0);
}

//***************** naprogramuje SPI FlashROM z daneho souboru **************
int spi_flash_program(char *file_name, DWord total_size)
{
  DWord address;
  FILE *fr=fopen(file_name,"rb");      // otevri soubor pro cteni
  if (fr==NULL)                        // pokud chyba, tak skonci
    {
    printf("ERROR: cannot open file %s\n",file_name);
    return(-1);
    }
  for (address=0; address<total_size; address+=SPI_FLASH_PAGE_SIZE)
    {                                  // pres vsechny stranky
    printf("Programming page at: %08lXh (%lu%%)\r",address,100*(address+SPI_FLASH_PAGE_SIZE)/total_size);
    spi_flash_wait_busy(1000);         // cekej dokud je nahozen busy flag
    spi_flash_write_enable();          // napred povol zapis, nuluje se automaticky po uspesnem zapisu stranky
    spi_do_buffer[0]=SPI_FLASH_PAGEPGM; // put command code to output buffer
    spi_do_buffer[1]=(address>>16)&0xFF;// put A23:16 to output buffer
    spi_do_buffer[2]=(address>>8)&0xFF; // put A15:8 to output buffer
    spi_do_buffer[3]=(address>>0)&0xFF; // put A7:0 to output buffer
    if (fread(spi_do_buffer+4,1,SPI_FLASH_PAGE_SIZE,fr)!=SPI_FLASH_PAGE_SIZE) // nacti stranku ze souboru
      {
      printf("ERROR: cannot read page %lu from file\n",address/SPI_FLASH_PAGE_SIZE);
      return(-2);
      }
    spi_rw_block(spi_do_buffer,spi_di_buffer,SPI_FLASH_PAGE_SIZE+4); // execute block read/write
    udelay(1000);                      // pockej chvilku
    }
  fclose(fr);                          // zavri soubor
  putchar('\n');
  return(0);
}

//***************** smaze vsechny protect bity, zpristupni zapis do SPI FlashROM (pokud to dovoli WP# pin)
int spi_flash_unlock(void)
{
  Byte status;
  spi_flash_write_enable();            // napred povol zapis, nuluje se automaticky po uspesnem zapisu do status registru
  status=spi_flash_status();           // zkontroluj status registr
  printf("Status = %02Xh, WE bit is %s\n",status,((status&0x2)==0) ? "disabled" : "enabled");
  if ((status&0x2)==0)                 // pokud se WE nenahodil
    return(-1);                        // tak skonci
  spi_do_buffer[0]=SPI_FLASH_WRSTAT;   // put command code to output buffer
  spi_do_buffer[1]=0;                  // status register = 0, all protections disabled
  spi_rw_block(spi_do_buffer,spi_di_buffer,2); // execute block read/write
  printf("Unlocking...\n");
  spi_flash_wait_busy(5000);           // cekej dokud se neprovede zapis
  status=spi_flash_status();           // zkontroluj status registr
  printf("Status = %02Xh, ",status);
  if ((status&0xBC)==0)                // pokud jsou protect bity nulove
    printf("device is unlocked now.\n");
  else
    printf("unlock failed, set WP# pin level high\n.");
  return(0);
}

//***************** smaze kompletne obsah SPI FlashROM (nastavi vsude FFh) **
int spi_flash_erase(void)
{
  Byte status;
  spi_flash_write_enable();            // napred povol zapis, nuluje se automaticky po uspesnem smazani
  status=spi_flash_status();           // zkontroluj status registr
  printf("Status = %02Xh, WE bit is %s\n",status,((status&0x2)==0) ? "disabled" : "enabled");
  if ((status&0x2)==0)                 // pokud se WE nenahodil
    return(-1);                        // tak skonci
  spi_rw_byte(SPI_FLASH_CHIPERASE);    // spust mazani cele FlashROM
  printf("Please wait...");
  spi_flash_wait_busy(20000);          // cekej dokud je nahozen busy flag
  printf(" done.\n");
  return(0);
}

//***************** precte a vypise JEDEC ID a typ pameti a pripadne nazev vyrobce a typ obvodu
int spi_flash_identify(void)
{
  DWord jid=spi_flash_read_jid();
  printf("SPI connected to LPT port at I/O base address: %Xh, SCK pulse width: t+%ldus\n",lpt_base,spi_sck_pulse_delay);
  printf("FlashROM JEDEC ID, type: %06lXh\n",jid);

  if (((jid>>16)&0xFF)==SPI_FLASH_JID_ATMEL) // Atmel FlashROM
    {
    printf("Atmel ");
    switch (jid&0xFFFF)
      {
      case 0x4400: printf("AT26DF041 (512kB)\n"); spi_flash_size=524288; break;
      case 0x4501: printf("AT26DF081A (1MB)\n"); spi_flash_size=1048576; break;
      case 0x4601: printf("AT26DF161A (2MB)\n"); spi_flash_size=2097152; break;
      case 0x4700: printf("AT26DF321 (4MB)\n"); spi_flash_size=4194304; break;
      default: printf("unknown chip\n"); return(0);
      }
    return(SPI_FLASH_JID_ATMEL);
    }

  if (((jid>>16)&0xFF)==SPI_FLASH_JID_MACRONIX) // Macronix FlashROM
    {
    printf("Macronix ");
    switch (jid&0xFFFF)
      {
      case 0x2011: printf("MX25L1005 (128kB)\n"); spi_flash_size=131072; break;
      case 0x2012: printf("MX25L2005 (256kB)\n"); spi_flash_size=262144; break;
      case 0x2013: printf("MX25L4005 (512kB)\n"); spi_flash_size=524288; break;
      case 0x2014: printf("MX25L8005 (1MB)\n"); spi_flash_size=1048576; break;
      case 0x2015: printf("MX25L1605 (2MB)\n"); spi_flash_size=2097152; break;
      case 0x2016: printf("MX25L3205 (4MB)\n"); spi_flash_size=4194304; break;
      case 0x2017: printf("MX25L6405 (8MB)\n"); spi_flash_size=8388608; break;
      default: printf("unknown chip\n"); return(0);
      }
    return(SPI_FLASH_JID_MACRONIX);
    }

  if (((jid>>16)&0xFF)==SPI_FLASH_JID_WINBOND) // Winbond FlashROM
    {
    printf("Winbond ");
    switch (jid&0xFFFF)
      {
      case 0x3011: printf("W25X10 (128kB)\n"); spi_flash_size=131072; break;
      case 0x3012: printf("W25X20 (256kB)\n"); spi_flash_size=262144; break;
      case 0x3013: printf("W25X40 (512kB)\n"); spi_flash_size=524288; break;
      case 0x3014: printf("W25X80 (1MB)\n"); spi_flash_size=1048576; break;
      case 0x3015: printf("W25X16 (2MB)\n"); spi_flash_size=2097152; break;
      case 0x3016: printf("W25X32 (4MB)\n"); spi_flash_size=4194304; break;
      case 0x3017: printf("W25X64 (8MB)\n"); spi_flash_size=8388608; break;
      default: printf("unknown chip\n"); return(0);
      }
    return(SPI_FLASH_JID_WINBOND);
    }

    if (((jid>>16)&0xFF)==SPI_FLASH_JID_SST) // SST FlashROM
    {
    printf("SST ");
    switch (jid&0xFFFF)
      {
      case 0x49BF: printf("SST25VF010 (128kB)\n"); spi_flash_size=131072; break;
      case 0x43BF: printf("SST25VF020 (256kB)\n"); spi_flash_size=262144; break;
      case 0x44BF: printf("SST25VF040 (512kB)\n"); spi_flash_size=524288; break;
      case 0x80BF: printf("SST25VF080 (1MB)\n"); spi_flash_size=1048576; break;
      case 0x2541: printf("SST25VF016 (2MB)\n"); spi_flash_size=2097152; break;
      case 0x2542: printf("SST25VF032 (4MB)\n"); spi_flash_size=4194304; break;
      case 0x2543: printf("SST25VF064 (8MB)\n"); spi_flash_size=8388608; break;
      case 0x2544: printf("SST25VF128 (16MB)\n"); spi_flash_size=16777216; break;
      case 0x258E: printf("SST25VF080B (1MB)\n"); spi_flash_size=1048576; break;
      case 0x2601: printf("SST26VF016 (2MB)\n"); spi_flash_size=2097152; break;
      case 0x2602: printf("SST26VF032 (4MB)\n"); spi_flash_size=4194304; break;
      case 0x2603: printf("SST26VF064 (8MB)\n"); spi_flash_size=8388608; break;
      default: printf("unknown chip\n"); return(0);
      }
    return(SPI_FLASH_JID_SST);
    }

  if (((jid>>16)&0xFF)==SPI_FLASH_JID_STM) // STM FlashROM
    {
    printf("ST Microelectronic ");
    switch (jid&0xFFFF)
      {
      case 0x2011: printf("M25P10 (128kB)\n"); spi_flash_size=131072; break;
      case 0x2012: printf("M25P20 (256kB)\n"); spi_flash_size=262144; break;
      case 0x2013: printf("M25P40 (512kB)\n"); spi_flash_size=524288; break;
      case 0x2014: printf("M25P80 (1MB)\n"); spi_flash_size=1048576; break;
      case 0x2015: printf("M25P16 (2MB)\n"); spi_flash_size=2097152; break;
      case 0x2016: printf("M25P32 (4MB)\n"); spi_flash_size=4194304; break;
      case 0x2017: printf("M25P64 (8MB)\n"); spi_flash_size=8388608; break;
      case 0x2018: printf("M25P128 (16MB)\n"); spi_flash_size=16777216; break;
      default: printf("unknown chip\n"); return(0);
      }
    return(SPI_FLASH_JID_STM);
    }
  printf("Unknown manufacturer\n");
  return(0);
}

//***************************************************************************
int spi_slow_program( char *filename )
{
  DWord address = 0L;
  FILE *fr=fopen(filename,"rb");      // otevri soubor pro cteni

  if( fr == NULL ) {                     // pokud chyba, tak skonci
    printf( "ERROR: cannot open file %s\n", filename );
    return( -1 );
  }
  printf( "Programming with %s\n", filename );
  while ( fread( spi_do_buffer, 1, 1, fr ) == 1 ) {
    if( ( address % 32768 ) == 0 )
      printf("\n%08lx ", address );
    if( ( address % 1024 ) == 0 )
      printf( "." );
    spi_rw_byte(SPI_FLASH_WREN);
    SPI_CLR_CS;
    spi_rw_byte_no_cs(SPI_FLASH_PAGEPGM);
    spi_rw_byte_no_cs(((address & 0xffffff) >> 16));
    spi_rw_byte_no_cs(((address & 0xffff) >> 8));
    spi_rw_byte_no_cs(address & 0xff);
    spi_rw_byte_no_cs((Byte) *spi_do_buffer);
    SPI_SET_CS;
    spi_rw_byte(SPI_FLASH_WRDIS);
    udelay(1000);
    address++;
  }
  printf( "\n" );
  
  fclose(fr);                          // zavri soubor

  return(0);
}

//***************************************************************************
int spi_program_test(DWord start, DWord count, Word pattern)
{
  printf("Starting at %ld for %ld bytes set to %02x\n", start, count, pattern);
  while( count-- )
  {
    if( ( address % 32768 ) == 0 )
      printf("\n%08lx ", address );
    if( ( address % 512 ) == 0 )
      printf( "." );
    spi_rw_byte(SPI_FLASH_WREN);
    SPI_CLR_CS;
    spi_rw_byte_no_cs(SPI_FLASH_PAGEPGM);
    spi_rw_byte_no_cs(((start & 0xffffff) >> 16));
    spi_rw_byte_no_cs(((start & 0xffff) >> 8));
    spi_rw_byte_no_cs(start & 0xff);
    spi_rw_byte_no_cs((Byte)pattern++);
    SPI_SET_CS;
    spi_rw_byte(SPI_FLASH_WRDIS);
    udelay(1000);
    start++;
  }
  printf( "\n" );

  return( 0 );
}

//***************** zpracuj pridavne volitelne parametry na konci ***********
void parse_options(int argc, int argnext, char *argv[])
{
  int i;
  for (i=argnext; i<argc; i++)         // pres vsechny dalsi argumenty do posledniho
    {
    if (tolower(argv[i][1])=='l')      // pokud je to /l - LPT IO base
      sscanf(argv[i]+2,"=%hi",&lpt_base);
    if (tolower(argv[i][1])=='d')      // pokud je to /d - SPI clock pulse delay
      sscanf(argv[i]+2,"=%lu",&spi_sck_pulse_delay);
    }
}

//***************** hlavni program ******************************************
int main(int argc, char *argv[])
{
  DWord count;
  DWord address;
  DWord start_address;
  DWord blk_size;
  Word pattern;
  Byte flash_status;
  Byte i;
  printf("\nSPI FlashROM Programmer 1.4 (C) 2008 by Martin Rehak; rayer@seznam.cz\n");
  printf("Compiled by GCC %s at %s, %s\n", __VERSION__, __TIME__, __DATE__);

  if (argc<2)                          // pokud na radce neni zadny parametr
    {
    printf("SYNTAX: spipgm /i|r|d|e|p|t|s [/l=iobase] [/d=delay]\n");
    printf("        /i - identify SPI FlashROM\n");
    printf("        /r [address] [size] - read & display data block (0x prefix = hexa number)\n");
    printf("        /d [filename] - dump entire FlashROM to file\n");
    printf("        /p [filename] - program entire FlashROM from file (without erase)\n");
    printf("        /e - erase entire FlashROM\n");
    printf("        /u - unlock write protection bits (may depend on WP# level)\n");
	printf("        /t [address] [size] [pattern] - fill data block with pattern\n");
	printf("        /s [filename] - program FlashROM with contents of file one byte at a time\n");
    printf("        /l=[port] LPT port I/O base address (default is 378h - LPT1)\n");
    printf("        /d=[usec] additional delay for SPI clock pulse width(default is 0)\n");
    printf("\nLPT to SPI pin assignment (based on BSD AVR programmer):\n");
    printf("pin  7 = D5 -> CS#\n");
    printf("pin  8 = D6 -> SCK (CLK)\n");
    printf("pin  9 = D7 -> MOSI (DIO)\n");
    printf("pin 10 = ACK -> MISO (DO)\n");
    printf("pin 18 = GND -> GND\n");
    return(1);
    }

  if (argc>1)                          // pokud je na radce alespon 1 parametr
    if (tolower(argv[1][1])=='i')      // pokud je prvni parametr i - identify
      {
      spi_init();                      // nastav linky SPI do vychoziho stavu
      parse_options(argc,2,argv);      // zpracuj volitelne parametry od 2.
      spi_flash_identify();            // identifikuj SPI FlashROM
      flash_status=spi_flash_status(); // precti status
      printf("Status = %02Xh (BPL, AAI, BP3, BP2, BP1, BP0, WEL, BSY)\n               ",flash_status);
      for (i=0; i<8; i++)              // pres vsechny bity statusu
        printf("%u    ",((flash_status<<i)&0x80)>>7); // vypis status binarne
      putchar('\n');
      if ((flash_status&0x80)!=0)      // check SRP bit
        printf("Status register is protected, you will need to set WP# high\n");
      }

  if (argc>3)                          // pokud jsou na radce alespon 3 parametry
    if (tolower(argv[1][1])=='r')      // pokud je prvni parametr r - read block data
      {
      spi_init();                      // nastav linky SPI do vychoziho stavu
      parse_options(argc,4,argv);      // zpracuj volitelne parametry od 4.
      sscanf(argv[2],"%li",&start_address); // precti pocatecni adresu
      sscanf(argv[3],"%li",&blk_size); // precti velikost bloku
      spi_flash_identify();            // identifikuj SPI FlashROM
      spi_flash_read_block(start_address,blk_size); // precti blok do bufferu
      if (blk_size<=(SPI_BUFFER_SIZE-4)) // pokud je velikost bloku mensi nez se vejde do bufferu
        spi_print_buffer(spi_di_buffer,blk_size,start_address); // vypis buffer na obrazovku
      }

  if (argc>2)                          // pokud jsou na radce alespon 2 parametry
    if (tolower(argv[1][1])=='d')      // pokud je prvni parametr d - dump FlashROM to file
      {
      spi_init();                      // nastav linky SPI do vychoziho stavu
      parse_options(argc,3,argv);      // zpracuj volitelne parametry od 3.
      if (spi_flash_identify()==0)     // identifikuj SPI FlashROM
        {                              // pokud se nedetekovala, zadej rucne velikost
        printf("Enter total FlashROM size in kB: ");
        scanf("%li",&spi_flash_size);
        spi_flash_size*=1024;          // precti velikost a preved na Byte
        }
      return(spi_flash_dump(argv[2],spi_flash_size)); // dumpuj pamet
      }

  if (argc>2)                          // pokud jsou na radce alespon 2 parametry
    if (tolower(argv[1][1])=='p')      // pokud je prvni parametr p - program FlashROM from file
      {
      spi_init();                      // nastav linky SPI do vychoziho stavu
      parse_options(argc,3,argv);      // zpracuj volitelne parametry od 3.
      if (spi_flash_identify()==0)     // identifikuj SPI FlashROM
        {                              // pokud se nedetekovala, zadej rucne velikost
        printf("Enter total FlashROM size in kB: ");
        scanf("%li",&spi_flash_size);
        spi_flash_size*=1024;          // precti velikost a preved na Byte
        }
      return(spi_flash_program(argv[2],spi_flash_size)); // programuj pamet
      }

  if (argc>1)                          // pokud je na radce alespon 1 parametr
    if (tolower(argv[1][1])=='e')      // pokud je prvni parametr e - chip erase
      {
      spi_init();                      // nastav linky SPI do vychoziho stavu
      parse_options(argc,2,argv);      // zpracuj volitelne parametry od 2.
      spi_flash_identify();            // identifikuj SPI FlashROM
      return(spi_flash_erase());       // vymaz kompletne obsah SPI FlashROM (nastavi bunky na FFh)
      }

  if (argc>1)                          // pokud je na radce alespon 1 parametr
    if (tolower(argv[1][1])=='u')      // pokud je prvni parametr u - protection unlock
      {
      spi_init();                      // nastav linky SPI do vychoziho stavu
      parse_options(argc,2,argv);      // zpracuj volitelne parametry od 2.
      spi_flash_identify();            // identifikuj SPI FlashROM
      return(spi_flash_unlock());      // vynuluj vsechny protection bity SPI FlashROM
      }

  if (argc>1)
    if(tolower(argv[1][1])=='t')
    {
      spi_init();
      parse_options(argc,5,argv);
      spi_flash_identify();            // identifikuj SPI FlashROM
      sscanf(argv[2],"%li", &address); // precti pocatecni adresu
      sscanf(argv[3],"%li", &count);   // precti velikost bloku
      sscanf(argv[4],"%x", &pattern);

      return(spi_program_test(address, count, pattern));
    }

  if (argc > 1)
    if (tolower(argv[1][1]=='s'))
    {
      spi_init();
      parse_options(argc, 2, argv );

      return(spi_slow_program(argv[2]));
    }

  return(0);
}
